"""
Module Unit test for data_analysis.py
"""

import unittest
from unittest import TestCase

import numpy as np
import pandas as pd

from eurybia.report.common import compute_col_types
from eurybia.report.data_analysis import perform_global_dataframe_analysis, perform_univariate_dataframe_analysis


class TestDataAnalysis(unittest.TestCase):
    """
    Unit test data_analysis.py
    """

    def test_perform_global_dataframe_analysis_1(self):
        """
        Unit test 1 perform_global_dataframe_analysis method
        """
        df = pd.DataFrame(
            {
                "string_data": ["a", "b", "c", "d", "e", np.nan],
                "bool_data": [True, True, False, False, False, np.nan],
                "int_data": [10, 20, 30, 40, 50, 0],
            }
        )

        dico = perform_global_dataframe_analysis(df)
        expected_d = {
            "number of features": "3",
            "number of observations": "6",
            "missing values": "2",
            "% missing values": "0.111",
        }
        TestCase().assertDictEqual(dico, expected_d)

    def test_perform_global_dataframe_analysis_2(self):
        """
        Unit test 2 perform_global_dataframe_analysis method
        """
        df = None
        dico = perform_global_dataframe_analysis(df)
        assert dico == {}

    def test_perform_univariate_dataframe_analysis_1(self):
        """
        Unit test 1 perform_univariate_dataframe_analysis method
        """
        df = pd.DataFrame(
            {
                "string_data": ["a", "b", "c", "d", "e", np.nan] * 10,
                "bool_data": [True, True, False, False, False, np.nan] * 10,
                "int_continuous_data": list(range(60)),
                "float_continuous_data": np.linspace(0, 2, 60),
                "int_cat_data": [1, 1, 1, 2, 2, 2] * 10,
                "float_cat_data": [0.2, 0.2, 0.2, 0.6, 0.6, 0.6] * 10,
            }
        )
        dico = perform_univariate_dataframe_analysis(df, col_types=compute_col_types(df))
        expected_d = {
            "int_continuous_data": {
                "count": "60",
                "mean": "29.5",
                "std": "17.5",
                "min": "0",
                "25%": "14.8",
                "50%": "29.5",
                "75%": "44.2",
                "max": "59",
            },
            "float_continuous_data": {
                "count": "60",
                "mean": "1",
                "std": "0.592",
                "min": "0",
                "25%": "0.5",
                "50%": "1",
                "75%": "1.5",
                "max": "2",
            },
            "int_cat_data": {"distinct values": "2", "missing values": "0"},
            "float_cat_data": {"distinct values": "2", "missing values": "0"},
            "string_data": {"distinct values": "5", "missing values": "10"},
            "bool_data": {"distinct values": "2", "missing values": "10"},
        }
        TestCase().assertDictEqual(dico, expected_d)

    def test_perform_univariate_dataframe_analysis_2(self):
        """
        Unit test 2 perform_univariate_dataframe_analysis method
        """
        df = None
        dico = perform_univariate_dataframe_analysis(df, col_types=compute_col_types(df))
        assert dico == {}



from typing import Any

report_text: dict[str, Any] = {
    "Index": {
        "01": "- Project information: report context and information",
        "02": "- Consistency Analysis: highlighting differences between the two datasets",
        "03": "- Data drift: In-depth data drift analysis",
        "04": "- Model drift: In-depth model drift analysis",
    },
    "Consistency analysis": {
        "01": (
            "The columns identified in this section have been automatically removed from this analysis. "
            "Their presence would always be sufficient for the datadrift classifier to perfectly discriminate "
            "the two datasets (maximal data drift, AUC=1)."
        ),
        "02": (
            "This section displays categorical features in which unique values differ. "
            "This analysis has been performed on unstratified samples of the baseline and current datasets. "
            "Missing or added unique values can be caused by this sampling. "
            "Columns identified in this section have been kept for the analysis."
        ),
    },
    "Data drift": {
        "01": (
            "The data drift detection methodology is based on the ability of a model classifier to identify "
            "whether a sample belongs to one or another dataset. For this purpose a target (0) is assigned "
            "to the baseline dataset and a second target (1) to the current dataset. "
            "A classification model (catboost) is trained to predict this target. As such, the data drift "
            "classifier performance is directly related to the difference between two datasets. A marked difference "
            " will lead to an easy classification (final AUC close to 1). Oppositely, highly similars datasets "
            " will lead to poor data drift classifier performance (final AUC close to 0.5)."
        ),
        "02": (
            "The closer your AUC is from 0.5 the less your data drifted. "
            "The closer your AUC is from 1 the more your data drifted."
        ),
        "03": (
            "Bar chart representing the feature importance of each feature for the datadrift classifier. "
            "This parameter is a direct measure of the importance of a feature to perform the classification."
        ),
        "04": (
            "Scatter plot depicting, for each feature, the feature importance of the deployed model "
            "as a function of the datadrift classifier "
            "feature importance. This graph thus highlight the real importance of a data drift "
            "for the deployed model classification. "
            "Interpretation based on graphical feature location:\n"
            "- Top left : Feature highly important for the deployed model and with low data drift\n"
            "- Bottom left : Feature with moderated importance for the deployed model and with low data drift.\n"
            "- Bottom right : Feature with moderated importance for the deployed model but with high data drift. "
            "This feature might require your attention.\n"
            "- Top right : Feature highly important for the deployed model and high drift. "
            "This feature requires your attention."
        ),
        "05": (
            "This section provides numerical and graphical analysis of the 2 datasets distributions, "
            "making easier the study of the most important variable for drift detection."
        ),
        "06": (
            "Histogram density showing the distributions of the production model outputs on "
            "both baseline and current datasets."
        ),
        "08": (
            "Jensen Shannon Divergence (JSD). "
            "The JSD measures the effect of a data drift on the deployed model performance. "
            "A value close to 0 indicates similar data distributions, while a value close to 1 "
            "tend to indicate distinct data distributions with a negative effect on the deployed model performance."
        ),
        "07": (
            "Bar chart showing the unique values distribution of a feature. "
            "Using the drop-down menu, it is possible to select the feature of interest. "
            "Features are sorted according to their respective importance in the datadrift classifier. "
            "For categorical features, the possible values are sorted by descending difference "
            "between the two datasets."
        ),
        "09": (
            "This graph represents the contribution of a variable to the data drift detection. "
            "This representation constitutes a support to understand the drift "
            "when the analysis of the dataset is unclear."
        ),
        "10": (
            "This graph represents the interactions between couple of variable to the data drift detection. "
            "This representation constitutes a support to understand the drift "
            "when the analysis of the dataset is unclear."
        ),
        "11": ("Line chart showing the metrics evolution of the datadrift classifier over the given period of time."),
    },
    "Model drift": {
        "01": (
            "This section provides support to monitor the production model's performance over time. "
            "This requires the performance history as input."
        ),
        "02": ("Line chart of deployed model performances as a function of time."),
    },
}

report_css: str = """
.bk-tab {
    background-color: white !important;
    font-size: 0.875rem;
    padding-top: 1em;
    padding-bottom: 1em;
    border-width: 3px !important;
    border-style: solid !important;
    border-color: white !important;
    color: #6b7280;
    margin-left: 1.5em;
    margin-right: 1.5em;
    text-align: left;
}

.bk-tab.bk-active {
    border-bottom-color: #4e46e5 !important;
    color: #101010;
}

.bk-header {
    position: sticky;
    top: 0;
    z-index: 100;
    width: 100%;
    background-color: white;
}

:host(.bk-above) .bk-tab:first-child {
    margin-left: 5rem !important;
}

.bk-above {
    width: 100%;
}

.bk-Row {
    width: 100%;
    display: grid;
    grid-template-columns: auto auto;
}

.bk-panel-models-layout-Column {
    margin-left: auto;
    margin-right: auto;
    width: 100%;
    max-width: 1280px;
    padding-left: 1rem;
    padding-right: 1rem;
}

.bk-panel-models-markup-HTML {
    width: 100%;
}

.bk-panel-models-plotly-PlotlyPlot {
    width: 100%;
    display: flex;
    justify-content: center;
}

h1 {
    font-size: 2.25rem;
    line-height: 1.33;
    font-weight: 700;
}

h2 {
    font-size: 1.5rem;
    line-height: 1.33;
    font-weight: 700;
}

h3 {
    font-size: 1.25rem;
    line-height: 1.6;
    font-weight: 600;
}

h4 {
    font-size: 1rem;
    line-height: 1.6;
    font-weight: 600;
}

ul {
    padding-left: 1.625em;
}

li {
    font-size: 1rem;
    line-height: 1.75;
    padding-left: 0.375em;
    margin-top: 0.5em;
    margin-bottom: 0.5em;
}

p {
    font-size: 1rem;
}

th {
    vertical-align: bottom;
    text-align: center !important;
    font-weight: 600;
    font-size: 1rem;
}

td {
    vertical-align: top !important;
    text-align: center !important;
    font-size: 1rem;
}

select {
    font-size: 1rem;
    font-weight: 700;
}


.hidden {
    display: none;
}

"""

report_jscallback: str = """
console.log("callback called");
var active_tab = active.active;
var top = document.getElementsByTagName('div')[1];
var elements = top.shadowRoot.children;
var current_tab = 0;
for (var i=0 ; i<elements.length ; i++) {
    var tagName = elements[i].tagName;
    if (tagName === 'DIV') {
        var className = elements[i].className;
        if (className !== "bk-header" && current_tab !== active_tab) {
            elements[i].style.removeProperty('visibility');
            elements[i].style['display'] = 'none';
        } else {
            elements[i].style.removeProperty("display");
        }
        current_tab = current_tab + 1;
    }
}
window.scrollTo(0, 0);
"""

select_callback: str = """
console.log("Tab = " + tab);
console.log("Key = " + key);
var f_class = this.value.replace(" ", "-").toLowerCase();
console.log("Feature = " + f_class);
var elts = document.querySelectorAll(".bk-above");
console.log("Elts : " + elts + " - length: " + elts.length);
for (let i=0; i<elts.length; i++) {
  var columns = elts[i].shadowRoot.querySelectorAll(tab);
  console.log("Columns : " + columns + " - length: " + columns.length);
  for (let j=0; j<columns.length; j++) {
    var nodes = columns[j].shadowRoot.querySelectorAll(key);
    console.log("Nodes : " + nodes + " - length: " + nodes.length);
    for (let k=0; k<nodes.length; k++) {
      if (nodes[k].classList.contains(f_class)) {
        nodes[k].classList.remove("hidden");
      } else {
        nodes[k].classList.add("hidden");
      }
    }
  }
}
"""



"""
Report generation helper module.
"""

from datetime import datetime
from typing import Optional

import pandas as pd
import panel as pn
from plotly.graph_objects import Violin
from shapash.explainer.smart_explainer import SmartExplainer

from eurybia import SmartDrift
from eurybia.report.project_report import DriftReport
from eurybia.report.properties import report_css, report_jscallback, report_text, select_callback

pn.extension("plotly")


def get_index_panel(dr: DriftReport, project_info_file: str, config_report: Optional[dict]) -> pn.Column:
    parts = []
    header_logo = pn.pane.PNG(
        "https://eurybia.readthedocs.io/en/latest/_images/eurybia-fond-clair.png?raw=true",
        styles={"max-width": "150px", "height": "auto"},
    )
    header_title = pn.pane.Markdown(f"# {dr.title_story}")
    header = pn.Row(header_logo, header_title)
    parts.append(header)

    if (
        config_report is not None
        and "title_description" in config_report.keys()
        and config_report["title_description"] != ""
    ):
        raw_title = config_report["title_description"]
        parts.append(pn.pane.Markdown(f"## {raw_title}"))

    content_parts = ["## Eurybia Report contents"]
    if project_info_file is not None:
        content_parts.append(report_text["Index"]["01"])
    content_parts.append(report_text["Index"]["02"])
    content_parts.append(report_text["Index"]["03"])
    if dr.smartdrift.data_modeldrift is not None:
        content_parts.append(report_text["Index"]["04"])
    content = pn.pane.Markdown("\n".join(content_parts))
    parts.append(content)

    # AUC
    auc_block = dr.smartdrift.plot.generate_indicator(
        fig_value=dr.smartdrift.auc, height=280, width=500, title="Datadrift classifier AUC"
    )
    auc_indicator = pn.pane.Plotly(auc_block)

    # Jensen-Shannon
    if dr.smartdrift.deployed_model is not None:
        JS_block = dr.smartdrift.plot.generate_indicator(
            fig_value=dr.smartdrift.js_divergence,
            height=280,
            width=500,
            title="Jensen Shannon Datadrift",
            min_gauge=0,
            max_gauge=0.2,
        )
        js_indicator = pn.pane.Plotly(JS_block)
        indicators = pn.Row(auc_indicator, js_indicator)

    else:
        indicators = pn.Row(auc_indicator)
    parts.append(indicators)

    return pn.Column(*parts, name="Index", css_classes=["index"])


def dict_to_text_blocks(text_dict: dict, level: int = 1) -> pn.Column:
    """
    This function recursively explores the dict and returns a Panel Column containing
    other groups and text blocks fed with the dict
    Parameters
    ----------
    text_dict: dict
        This dict must contain string as keys, and dicts or strings as values
    level: int = 1
        Recursion level, starting at 1 to allow for direct string manipulation
    Returns
    ----------
    pn.Column
        Column of blocks
    """
    blocks = []
    text = ""
    for k, v in text_dict.items():
        if isinstance(v, (str, int, float)) or v is None:
            if k.lower() == "date" and isinstance(v, str) and v.lower() == "auto":
                v = str(datetime.now())[:-7]
            text += f"**{k}** : {v}  \n"
        elif isinstance(v, dict):
            if text != "":
                blocks.append(pn.pane.Markdown(text))
                text = ""
            blocks.append(
                pn.Column(pn.pane.Markdown("#" * min(level, 6) + " " + str(k)), dict_to_text_blocks(v, level + 1))
            )
    if text != "":
        blocks.append(pn.pane.Markdown(text))
    return pn.Column(*blocks)


def get_project_information_panel(dr: DriftReport) -> Optional[pn.Column]:
    if dr.metadata is None:
        return None
    blocks = dict_to_text_blocks(dr.metadata)
    return pn.Column(*blocks, name="Project information", styles=dict(display="none"))


def get_consistency_analysis_panel(dr: DriftReport) -> pn.Column:
    # Title
    blocks = [pn.pane.Markdown("# Consistency Analysis")]

    # Manually ignored coluumns
    ignore_cols = pd.DataFrame({"ignore_cols": dr.smartdrift.ignore_cols}).rename(
        columns={"ignore_cols": "Ignored columns"}
    )
    blocks += [
        pn.pane.Markdown("## Ignored columns in the report (manually excluded)"),
    ]
    if len(ignore_cols) > 0:
        blocks += [pn.pane.DataFrame(ignore_cols)]
    else:
        blocks += [pn.pane.Markdown("- Ignored columns : None.")]

    # Column mismatches
    blocks += [
        pn.pane.Markdown("## Consistency checks: column match between the 2 datasets."),
        pn.pane.Markdown(report_text["Consistency analysis"]["01"]),
    ]
    for k, v in dr.smartdrift.pb_cols.items():
        if len(v) > 0:
            blocks += [pn.pane.DataFrame(pd.DataFrame(v).transpose())]
        else:
            blocks += [pn.pane.Markdown(f"- No {k.lower()} have been detected.")]

    blocks += [
        pn.pane.Markdown("###  Unique values identified"),
        pn.pane.Markdown(report_text["Consistency analysis"]["02"]),
    ]
    if len(dr.smartdrift.err_mods) > 0:
        blocks += [
            pn.pane.DataFrame(
                pd.DataFrame(dr.smartdrift.err_mods)
                .rename(columns={"err_mods": "Modalities present in one dataset and absent in the other :"})
                .transpose(),
            )
        ]
    else:
        blocks += [
            pn.pane.Markdown("- No modalities have been detected as present in one dataset and absent in the other.")
        ]

    return pn.Column(*blocks, name="Consistency Analysis", styles=dict(display="none"), css_classes=["information"])


def get_select_plots(labels: list, key: str, tab: str, figures: list) -> list:
    blocks = []
    select = pn.widgets.Select(value=labels[0], options=labels)
    select.jscallback(args={"key": f".{key}", "tab": tab}, value=select_callback)
    blocks += [select]
    for i in range(len(labels)):
        f_class = labels[i].replace(" ", "-").lower()
        css_classes = [f_class, key]
        if labels[i] != labels[0]:
            css_classes.append("hidden")
        for figure_trace in figures[i].data:
            if isinstance(figure_trace, Violin):
                figure_trace.update(side="both")
        figures[i].update_layout(width=1240)
        node = pn.pane.Plotly(figures[i], name=labels[i], css_classes=css_classes)
        blocks += [node]
    return blocks


def get_select_tables(labels: list, key: str, tab: str, tables: list) -> list:
    blocks = []
    select = pn.widgets.Select(value=labels[0], options=labels)
    select.jscallback(args={"key": f".{key}", "tab": tab}, value=select_callback)
    blocks += [select]
    for i in range(len(labels)):
        f_class = labels[i].replace(" ", "-").lower()
        css_classes = [f_class, key]
        if i > 0:
            css_classes.append("hidden")
        node = pn.pane.DataFrame(tables[i], css_classes=css_classes)
        blocks += [node]
    return blocks


def get_data_drift_panel(dr: DriftReport) -> pn.Column:
    blocks = [
        pn.pane.Markdown("# Data drift"),
        pn.pane.Markdown(report_text["Data drift"]["01"]),
        pn.pane.Markdown("## Detecting data drift"),
        pn.pane.Markdown("### Datadrift classifier model perfomances"),
        pn.pane.Markdown(report_text["Data drift"]["02"]),
    ]
    auc = dr.smartdrift.plot.generate_indicator(
        fig_value=dr.smartdrift.auc, height=300, width=500, title="Datadrift classifier AUC"
    )
    blocks += [pn.pane.Plotly(auc)]

    blocks += [
        pn.pane.Markdown("## Importance of features in data drift"),
        pn.pane.Markdown("### Global feature importance plot"),
        pn.pane.Markdown(report_text["Data drift"]["03"]),
    ]

    fig_features_importance = dr.explainer.plot.features_importance()
    fig_features_importance.update_layout(width=1240)
    blocks += [pn.pane.Plotly(fig_features_importance)]
    if dr.smartdrift.deployed_model is not None:
        fig_scatter_feature_importance = dr.smartdrift.plot.scatter_feature_importance()
        fig_scatter_feature_importance.update_layout(width=1240)
        blocks += [
            pn.pane.Markdown("### Feature importance overview"),
            pn.pane.Markdown(report_text["Data drift"]["04"]),
            pn.pane.Plotly(fig_scatter_feature_importance),
        ]

    blocks += [
        pn.pane.Markdown("## Dataset analysis"),
        pn.pane.Markdown(report_text["Data drift"]["05"]),
        pn.pane.Markdown("### Global analysis"),
        pn.pane.DataFrame(dr._display_dataset_analysis_global()),
        pn.pane.Markdown("### Univariate analysis"),
        pn.pane.Markdown(report_text["Data drift"]["07"]),
    ]

    distribution_figures, labels, distribution_tables = dr.display_dataset_analysis(global_analysis=False)["univariate"]
    distribution_plots_blocks = get_select_plots(
        labels=labels, key="distribution-plot", tab=".data-drift", figures=distribution_figures
    )
    blocks += distribution_plots_blocks
    distribute_tables_blocks = get_select_tables(
        labels=labels, key="distribution-table", tab=".data-drift", tables=distribution_tables
    )
    blocks += distribute_tables_blocks

    if dr.smartdrift.deployed_model is not None:
        fig_01 = dr.smartdrift.plot.generate_fig_univariate(df_all=dr.smartdrift.df_predict, col="Score", hue="dataset")
        fig_01.update_layout(width=1240)
        blocks += [
            pn.pane.Markdown("### Distribution of predicted values"),
            pn.pane.Markdown(report_text["Data drift"]["06"]),
            pn.pane.Plotly(fig_01),
            pn.pane.Markdown(report_text["Data drift"]["08"]),
        ]
        js_fig = dr.smartdrift.plot.generate_indicator(
            fig_value=dr.smartdrift.js_divergence,
            height=280,
            width=500,
            title="Jensen Shannon Datadrift",
            min_gauge=0,
            max_gauge=0.2,
        )
        blocks += [pn.pane.Plotly(js_fig)]

    contribution_figures, contribution_labels = dr.display_model_contribution()

    blocks += [
        pn.pane.Markdown("## Feature contribution on data drift's detection"),
        pn.pane.Markdown(report_text["Data drift"]["09"]),
    ]
    contribution_plots_blocks = get_select_plots(
        labels=contribution_labels,
        key="contribution-plot",
        tab=".data-drift",
        figures=contribution_figures,
    )
    blocks += contribution_plots_blocks

    fig_02 = dr.explainer.plot.top_interactions_plot(nb_top_interactions=10)
    fig_02.update_layout(width=1240)
    blocks += [
        pn.pane.Markdown("## Feature interaction on data drift's detection"),
        pn.pane.Markdown(report_text["Data drift"]["10"]),
        pn.pane.Plotly(fig_02),
    ]

    if dr.smartdrift.historical_auc is not None:
        fig = dr.smartdrift.plot.generate_historical_datadrift_metric()
        fig.update_layout(width=1240)
        blocks += [
            pn.pane.Markdown("## Historical Data drift"),
            pn.pane.Markdown(report_text["Data drift"]["11"]),
            pn.pane.Plotly(fig),
        ]

    return pn.Column(*blocks, name="Data drift", styles=dict(display="none"), css_classes=["data-drift"])


def get_model_drift_panel(dr: DriftReport) -> pn.Column:
    """
    This function generates and returns a Panel Column page containing the Eurybia model drift analysis

    Parameters
    ----------
    dr : DriftReport
        DriftReport object

    Returns
    ----------
    pn.Column
    """

    blocks = [
        pn.pane.Markdown("# Model drift"),
        pn.pane.Markdown(report_text["Model drift"]["01"]),
        pn.pane.Markdown("## Performance evolution of the deployed model"),
        pn.pane.Markdown(report_text["Model drift"]["02"]),
    ]

    if dr.smartdrift.data_modeldrift is None:
        blocks += [pn.pane.Markdown("## Smartdrift.data_modeldrift is None")]
    else:
        figures, labels = dr.display_data_modeldrift()
        if labels == []:
            figures[0].update_layout(width=1240)
            blocks += [pn.pane.Plotly(figures[0])]
        else:
            list_blocks = get_select_plots(labels=labels, key="modeldrift-plot", tab=".model-drift", figures=figures)
            blocks += list_blocks

    return pn.Column(*blocks, name="Model drift", styles=dict(display="none"), css_classes=["model-drift"])


def execute_report(
    smartdrift: SmartDrift,
    explainer: SmartExplainer,
    project_info_file: str,
    output_file: str,
    config_report: Optional[dict] = {},
) -> None:
    """
    Creates the report

    Parameters
    ----------
    smartdrift : eurybia.core.smartdrift.SmartDrift object
        Compiled SmartDrift class
    explainer : shapash.explainer.smart_explainer.SmartExplainer object
        Compiled shapash explainer.
    project_info_file : str
        Path to the file used to display some information about the project in the report.
    config_report : dict, optional
        Report configuration options.
    output_file : str
            Path to the HTML file to write
    """
    dr = DriftReport(
        smartdrift=smartdrift,
        explainer=explainer,
        project_info_file=project_info_file,
        config_report=config_report,
    )

    tab_list = []
    tab_list.append(get_index_panel(dr, project_info_file, config_report))
    if project_info_file is not None:
        tab_list.append(get_project_information_panel(dr))
    tab_list.append(get_consistency_analysis_panel(dr))
    tab_list.append(get_data_drift_panel(dr))
    if dr.smartdrift.data_modeldrift is not None:
        tab_list.append(get_model_drift_panel(dr))

    pn.config.raw_css.append(report_css)
    report = pn.Tabs(*tab_list, css_classes=["main-report"])
    report.jscallback(args={"active": report}, active=report_jscallback)
    report.save(output_file, embed=True)



"""
Unit test of style_utils
"""
import re
import unittest

from eurybia.style.style_utils import colors_loading, define_style, select_palette


class TestStyle_utils(unittest.TestCase):
    """
    Class of Unit test for style_utils
    """

    def rgb_string_detector(self, string_val):
        """
        check rgb() or rgba() format of a str variable
        """
        matching = re.match(r"^rgba?\((\d+),\s*(\d+),\s*(\d+)(?:,\s*(\d+(?:\.\d+)?))?\)$", string_val)
        matching = False if matching is None else True
        return matching

    def test_colors_loading(self):
        """
        test of colors_loading
        """
        all_colors = colors_loading()
        for palette in all_colors.keys():
            for cle in all_colors[palette].keys():
                entries = all_colors[palette][cle]
                if isinstance(entries, dict):
                    check_list = list(entries.values())
                elif isinstance(entries, str):
                    check_list = [entries]
                elif isinstance(entries, list):
                    check_list = [x[1] for x in entries] if isinstance(entries[0], list) else entries
                for colors in check_list:
                    if self.rgb_string_detector(colors) is False:
                        print(colors)
                    assert self.rgb_string_detector(colors) is True

    def test_select_palette(self):
        """
        test of select_palette
        """
        available_palettes = colors_loading()
        default_dict = select_palette(available_palettes, "eurybia")
        assert len(list(default_dict.keys())) > 0
        with self.assertRaises(ValueError):
            select_palette(available_palettes, "ERROR_palette")

    def test_define_style(self):
        """
        test of define_style : check that each entry in json file is ok to define style_dict
        """
        available_palettes = colors_loading()
        for palette_name in available_palettes.keys():
            palette = select_palette(available_palettes, palette_name)
            style_dict = define_style(palette)
            assert len(list(style_dict.keys())) > 0



"""
Unit tests for SmartPlotter.py
"""

import copy
import unittest
from os import path
from pathlib import Path
from unittest.mock import patch

import numpy as np
import pandas as pd
import plotly
from sklearn.model_selection import train_test_split

from eurybia.core.smartdrift import SmartDrift
from eurybia.report.common import VarType
from eurybia.style.style_utils import colors_loading, select_palette

TITANIC_PATH = "eurybia/data/titanicdata.csv"


class TestSmartPlotter(unittest.TestCase):
    """
    Unit test SmartPlotter.py
    """

    def setUp(self):
        """SetUp"""

        script_path = Path(path.abspath(__file__)).parent.parent.parent.parent
        titanic_csv = path.join(script_path, TITANIC_PATH)
        titanic_df = pd.read_csv(titanic_csv)
        titanic_df_1, titanic_df_2 = train_test_split(titanic_df, test_size=0.5, random_state=42)
        self.smartdrift = SmartDrift(titanic_df_1, titanic_df_2)

    @patch("eurybia.core.smartplotter.SmartPlotter.generate_fig_univariate_continuous")
    @patch("eurybia.core.smartplotter.SmartPlotter.generate_fig_univariate_categorical")
    def test_generate_fig_univariate_1(self, mock_plot_cat, mock_plot_cont):
        """
        Unit test for generate_fig_univariate()'s method
        """
        df = pd.DataFrame(
            {
                "string_data": ["a", "b", "c", "d", "e", np.nan],
                "data_train_test": ["train", "train", "train", "train", "test", "test"],
            }
        )
        dict_color_palette = {
            "df_bBaseline": (74 / 255, 99 / 255, 138 / 255, 0.7),
            "df_Current": (244 / 255, 192 / 255, 0),
            "true": (74 / 255, 99 / 255, 138 / 255, 0.7),
            "pred": (244 / 255, 192 / 255, 0),
            "train": (74 / 255, 99 / 255, 138 / 255, 0.7),
            "test": (244 / 255, 192 / 255, 0),
            "df_baseline": (74 / 255, 99 / 255, 138 / 255, 0.7),
            "df_current": (244 / 255, 192 / 255, 0),
        }

        self.smartdrift.plot.generate_fig_univariate(
            "string_data", "data_train_test", df, dict_color_palette=dict_color_palette
        )
        mock_plot_cat.assert_called_once()
        self.assertEqual(mock_plot_cont.call_count, 0)

    @patch("eurybia.core.smartplotter.SmartPlotter.generate_fig_univariate_continuous")
    @patch("eurybia.core.smartplotter.SmartPlotter.generate_fig_univariate_categorical")
    def test_generate_fig_univariate_2(self, mock_plot_cat, mock_plot_cont):
        """
        Unit test for generate_fig_univariate()'s method
        """
        df = pd.DataFrame(
            {"int_data": list(range(50)), "data_train_test": ["train", "train", "train", "train", "test"] * 10}
        )

        dict_color_palette = {
            "df_Baseline": (74 / 255, 99 / 255, 138 / 255, 0.7),
            "df_Current": (244 / 255, 192 / 255, 0),
            "true": (74 / 255, 99 / 255, 138 / 255, 0.7),
            "pred": (244 / 255, 192 / 255, 0),
            "train": (74 / 255, 99 / 255, 138 / 255, 0.7),
            "test": (244 / 255, 192 / 255, 0),
            "df_baseline": (74 / 255, 99 / 255, 138 / 255, 0.7),
            "df_current": (244 / 255, 192 / 255, 0),
        }

        self.smartdrift.plot.generate_fig_univariate(
            "int_data", "data_train_test", df, dict_color_palette=dict_color_palette
        )
        mock_plot_cont.assert_called_once()
        self.assertEqual(mock_plot_cat.call_count, 0)

    @patch("eurybia.core.smartplotter.SmartPlotter.generate_fig_univariate_continuous")
    @patch("eurybia.core.smartplotter.SmartPlotter.generate_fig_univariate_categorical")
    def test_generate_fig_univariate_3(self, mock_plot_cat, mock_plot_cont):
        """
        Unit test for generate_fig_univariate()'s method
        """
        df = pd.DataFrame(
            {
                "int_cat_data": [10, 10, 20, 20, 20, 10],
                "data_train_test": ["train", "train", "train", "train", "test", "test"],
            }
        )
        dict_color_palette = {
            "df_Baseline": (74 / 255, 99 / 255, 138 / 255, 0.7),
            "df_Current": (244 / 255, 192 / 255, 0),
            "true": (74 / 255, 99 / 255, 138 / 255, 0.7),
            "pred": (244 / 255, 192 / 255, 0),
            "train": (74 / 255, 99 / 255, 138 / 255, 0.7),
            "test": (244 / 255, 192 / 255, 0),
            "df_baseline": (74 / 255, 99 / 255, 138 / 255, 0.7),
            "df_current": (244 / 255, 192 / 255, 0),
        }

        self.smartdrift.plot.generate_fig_univariate(
            "int_cat_data", "data_train_test", df, dict_color_palette=dict_color_palette
        )
        mock_plot_cat.assert_called_once()
        self.assertEqual(mock_plot_cont.call_count, 0)

    def test_generate_fig_univariate_continuous(self):
        """
        Unit test for generate_fig_univariate_continuous()'s method
        """
        df = pd.DataFrame(
            {
                "int_data": [10, 20, 30, 40, 50, 0],
                "data_train_test": ["train", "train", "train", "train", "test", "test"],
            }
        )
        dict_color_palette = {
            "df_Baseline": (74 / 255, 99 / 255, 138 / 255, 0.7),
            "df_Current": (244 / 255, 192 / 255, 0),
            "true": (74 / 255, 99 / 255, 138 / 255, 0.7),
            "pred": (244 / 255, 192 / 255, 0),
            "train": (74 / 255, 99 / 255, 138 / 255, 0.7),
            "test": (244 / 255, 192 / 255, 0),
            "df_baseline": (74 / 255, 99 / 255, 138 / 255, 0.7),
            "df_current": (244 / 255, 192 / 255, 0),
        }
        fig = self.smartdrift.plot.generate_fig_univariate_continuous(
            df, "int_data", "data_train_test", dict_color_palette=dict_color_palette
        )
        assert isinstance(fig, plotly.graph_objs._figure.Figure)

    def test_generate_fig_univariate_categorical(self):
        """
        Unit test for generate_fig_univariate_categorical()'s method
        """
        df = pd.DataFrame(
            {"int_data": [0, 0, 0, 1, 1, 0], "data_train_test": ["train", "train", "train", "train", "test", "test"]}
        )
        dict_color_palette = {
            "df_Baseline": (74 / 255, 99 / 255, 138 / 255, 0.7),
            "df_Current": (244 / 255, 192 / 255, 0),
            "true": (74 / 255, 99 / 255, 138 / 255, 0.7),
            "pred": (244 / 255, 192 / 255, 0),
            "train": (74 / 255, 99 / 255, 138 / 255, 0.7),
            "test": (244 / 255, 192 / 255, 0),
            "df_baseline": (74 / 255, 99 / 255, 138 / 255, 0.7),
            "df_current": (244 / 255, 192 / 255, 0),
        }

        fig = self.smartdrift.plot.generate_fig_univariate_categorical(
            df, "int_data", "data_train_test", dict_color_palette=dict_color_palette
        )
        nb_bars = len(fig.to_dict()["data"][0]["y"]) + len(fig.to_dict()["data"][1]["y"])
        assert nb_bars == 4  # Number of bars

    def test_scatter_feature_importance(self):
        """
        Unit test for scatter_feature_importance()'s method
        """
        importances = {
            "feature": ["Parch", "Embarked", "SibSp", "Age", "Fare", "Pclass", "Title", "Sex"],
            "deployed_model": [0.014573, 0.046866, 0.054405, 0.068945, 0.102350, 0.155283, 0.255974, 0.263060],
            "datadrift_classifier": [0.045921, 0.118159, 0.018614, 0.137789, 0.184038, 0.143481, 0.000000, 0.162925],
        }
        features_importances = pd.DataFrame(importances)
        stat_test = {
            "feature": ["Parch", "Embarked", "SibSp", "Age", "Fare", "Pclass", "Title", "Sex"],
            "testname": [
                "Chi-Square",
                "Chi-Square",
                "K-Smirnov",
                "Chi-Square",
                "Chi-Square",
                "K-Smirnov",
                "Chi-Square",
                "Chi-Square",
            ],
            "statistic": [
                2.6092852997885805,
                1.0551275457681237,
                0.05108076787423792,
                11.737348438770884,
                3.7354255436570627,
                0.05752002821585126,
                3.0136100828147927,
                13.864955367956966,
            ],
            "pvalue": [
                0.2712694558959706,
                0.30432911545137264,
                0.5753367740062436,
                0.06809137774132133,
                0.7124287331705583,
                0.4263491432475492,
                0.22161690489144445,
                0.6087737509479118,
            ],
        }
        statistical_tests = pd.DataFrame(stat_test)

        fig = self.smartdrift.plot.scatter_feature_importance(
            feature_importance=features_importances.set_index("feature"),
            datadrift_stat_test=statistical_tests.set_index("feature"),
        )
        assert isinstance(fig, plotly.graph_objs._figure.Figure)
        assert len(fig.data) == 1
        assert len(fig.data[0]["x"]) == features_importances.shape[0]

    def test_generate_historical_datadrift_metric_1(self):
        """
        Unit test for generate_historical_datadrift_metric()'s method
        """
        auc = {"date": ["23/09/2021", "23/08/2021", "23/07/2021"], "auc": [0.528107, 0.532106, 0.510653]}
        datadrift_historical = pd.DataFrame(auc)
        fig = self.smartdrift.plot.generate_historical_datadrift_metric(datadrift_historical)
        assert isinstance(fig, plotly.graph_objs._figure.Figure)

    def test_generate_historical_datadrift_metric_2(self):
        """
        Unit test for generate_historical_datadrift_metric()'s method
        """
        auc = {
            "date": ["23/09/2021", "23/08/2021", "23/07/2021"],
            "auc": [0.528107, 0.532106, 0.510653],
            "JS_predict": [0.28107, 0.32106, 0.50653],
        }
        datadrift_historical = pd.DataFrame(auc)
        fig = self.smartdrift.plot.generate_historical_datadrift_metric(datadrift_historical)
        assert isinstance(fig, plotly.graph_objs._figure.Figure)

    def test_generate_modeldrift_data_1(self):
        """
        Unit test for generate_modeldrift_data()'s method
        """
        lift_info = {
            "Date": ["01/07/2019", "01/08/2019"],
            "plage_historique": [1, 1],
            "target": ["souscription", "souscription"],
            "type_lift": ["lift10", "lift10"],
            "lift": [3.32, 3.18],
        }
        df_lift = pd.DataFrame(lift_info)

        fig = self.smartdrift.plot.generate_modeldrift_data(df_lift, metric="lift")
        assert isinstance(fig, plotly.graph_objs._figure.Figure)

    def test_generate_modeldrift_data_2(self):
        """
        Unit test for generate_modeldrift_data()'s method
        """
        lift_info = {
            "Date": ["01/07/2019", "01/08/2019"],
            "plage_historique": [1, 1],
            "target": ["souscription", "souscription"],
            "type_lift": ["lift10", "lift10"],
            "lift": [3.32, 3.18],
        }
        df_lift = pd.DataFrame(lift_info)

        fig = self.smartdrift.plot.generate_modeldrift_data(
            df_lift, metric="lift", reference_columns=["plage_historique", "target", "type_lift"]
        )
        assert isinstance(fig, plotly.graph_objs._figure.Figure)

    def test_define_style_attributes(self):
        # clear style attributes
        del self.smartdrift.plot._style_dict

        colors_dict = copy.deepcopy(select_palette(colors_loading(), "eurybia"))
        self.smartdrift.plot.define_style_attributes(colors_dict=colors_dict)

        assert hasattr(self.smartdrift.plot, "_style_dict")
        assert len(list(self.smartdrift.plot._style_dict.keys())) > 0

    def test_generate_indicator(self):
        """
        Unit test for generate_indicator()'s method
        """
        self.smartdrift.compile()
        fig = self.smartdrift.plot.generate_indicator(fig_value=self.smartdrift.auc, height=300, width=500, title="AUC")
        assert isinstance(fig, plotly.graph_objs._figure.Figure)



"""
Detection Drift Model Module
"""

catboost_hyperparameter_type = {
    "max_depth": ["<class 'int'>"],
    "l2_leaf_reg": ["<class 'float'>", "<class 'int'>"],
    "learning_rate": ["<class 'float'>"],
    "iterations": ["<class 'int'>"],
    "use_best_model": ["<class 'bool'>"],
    "custom_loss": ["<class 'str'>", "<class 'bool'>"],
    "loss_function": ["<class 'str'>", "<class 'object'>"],
    "eval_metric": ["<class 'str'>", "<class 'object'>"],
}

catboost_hyperparameter_init = {
    "max_depth": 8,
    "l2_leaf_reg": 19,
    "learning_rate": 0.166905,
    "iterations": 250,
    "use_best_model": True,
    "custom_loss": ["Accuracy", "AUC", "Logloss"],
    "loss_function": "Logloss",
    "eval_metric": "Logloss",
}



"""
Common functions used in report
"""

import os
from enum import Enum
from numbers import Number
from typing import Callable, Dict, Optional, Union

import pandas as pd

# from pandas.api.types import is_bool_dtype, is_categorical_dtype, is_numeric_dtype, is_string_dtype
from pandas.api.types import infer_dtype, is_numeric_dtype


class VarType(Enum):
    """
    Helper class to indicate the type of a variable.
    """

    TYPE_CAT = "Categorical"
    TYPE_NUM = "Numeric"
    TYPE_UNSUPPORTED = "Unsupported"

    def __str__(self) -> str:
        return str(self.value)


def display_value(value: float, thousands_separator: str = ",", decimal_separator: str = ".") -> str:
    """
    Display a value as a string with specific format.
    Parameters
    ----------
    value : float
        Value to display.
    thousands_separator : str
        The separator used to separate thousands.
    decimal_separator : str
        The separator used to separate decimal values.
    Returns
    -------
    str
    Examples
    --------
    >>> display_value(1255000, thousands_separator=',')
    '1,255,000'
    """
    value_str = f"{value:,}".replace(",", "/thousands/").replace(".", "/decimal/")
    return value_str.replace("/thousands/", thousands_separator).replace("/decimal/", decimal_separator)


def replace_dict_values(obj: Dict, replace_fn: Callable, *args) -> dict:
    """
    Recursively iterates over all values of obj and changes its values using the replace_fn
    Parameters
    ----------
    obj : dict
    replace_fn : callable
    Returns
    -------
    dict
    """
    for k, value in obj.items():
        if isinstance(value, dict):
            obj[k] = replace_dict_values(value, replace_fn)
        elif isinstance(value, Number):
            obj[k] = replace_fn(value, *args)
    return obj


def series_dtype(s: pd.Series) -> VarType:
    """
    Computes the type of a pandas series.
    Parameters
    ----------
    s : pd.Series
        The series for which we wish to determine the type.
    Returns
    -------
    VarType
    """
    if infer_dtype(s) == "boolean":
        return VarType.TYPE_CAT
    elif infer_dtype(s, skipna=True) == "string":
        return VarType.TYPE_CAT
    elif isinstance(s.dtype, pd.CategoricalDtype):
        return VarType.TYPE_CAT
    elif is_numeric_dtype(s):
        if numeric_is_continuous(s):
            return VarType.TYPE_NUM
        else:
            return VarType.TYPE_CAT
    else:
        return VarType.TYPE_UNSUPPORTED


def numeric_is_continuous(s: pd.Series):
    """
    Function that returns True if a
    numeric pandas series is continuous and False if it is categorical.
    Parameters
    ----------
    s : pd.Series
    Returns
    -------
    bool
    """
    # This test could probably be improved
    n_unique = s.nunique()
    return True if n_unique > 15 else False


def get_callable(path: str):
    """
    This function is similar to the _locate function in Hydra library
    Locate an object by name or dotted path, importing as necessary.
    """
    if path == "":
        raise ImportError("Empty path")
    import builtins
    from importlib import import_module

    parts = [part for part in path.split(".") if part]
    module = None
    for number in reversed(range(len(parts))):
        try:
            mod = ".".join(parts[:number])
            module = import_module(mod)
        except Exception as e:
            if number == 0:
                raise ImportError(f"Error loading module '{path}'") from e
            continue
        if module:
            break
    if module:
        obj = module
    else:
        obj = builtins
    for part in parts[number:]:
        mod = mod + "." + part
        if not hasattr(obj, part):
            try:
                import_module(mod)
            except Exception as e:
                raise ImportError(f"Encountered error: `{e}` when loading module '{path}'") from e
        obj = getattr(obj, part)
    if isinstance(obj, type):
        obj_type: type = obj
        return obj_type
    elif callable(obj):
        obj_callable = obj
        return obj_callable
    else:
        # dummy case
        raise ValueError(f"Invalid type ({type(obj)}) found for {path}")


def compute_col_types(df_all: Optional[pd.DataFrame]) -> Dict:
    """
    Computes the type of each column and stores the result in a dict.
    Parameters
    ----------
    df_all : pd.DataFrame, optional
    Returns
    -------
    col_types : dict
        The types of each column
    """
    if df_all is None:
        return {}
    return {col: series_dtype(df_all[col]) for col in df_all.columns}


def load_saved_df(path: str) -> Union[pd.DataFrame, None]:
    """
    Loads a pandas DataFrame that was saved using pd.to_csv method.
    Parameters
    ----------
    path : str
        Path to the dataframe object
    Returns
    -------
    pd.DataFrame or None
    """
    if os.path.exists(path):
        return pd.read_pickle(path)

    else:
        return None



"""
Unit tests for SmartDrift
"""
import os
import unittest
from os import path
from pathlib import Path
from unittest.mock import Mock, patch

import numpy as np
import pandas as pd
import pytest
import shapash
from category_encoders import OrdinalEncoder
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.model_selection import train_test_split

from eurybia import SmartDrift
from eurybia.style.style_utils import colors_loading, select_palette

TITANIC_PATH = "eurybia/data/titanicdata.csv"
TITANIC_ORIGINAL_PATH = "eurybia/data/titanic_original.csv"
TITANIC_ALTERED_PATH = "eurybia/data/titanic_altered.csv"
PROJECT_INFO_PATH = "eurybia/data/project_info_titanic.yml"
REPORT_PATH = "./test_report.html"


class TestSmartDrift(unittest.TestCase):
    """
    Unit test on SmartDrift class
    """

    def setUp(self):
        """SetUp"""
        script_path = Path(path.abspath(__file__)).parent.parent.parent.parent
        titanic_csv = path.join(script_path, TITANIC_PATH)
        titanic_df = pd.read_csv(titanic_csv)
        del titanic_df["Name"]
        titanic_df = titanic_df.drop("PassengerId", axis=1)
        titanic_df_1, titanic_df_2 = train_test_split(titanic_df, test_size=0.5, random_state=42)
        titanic_df_1 = titanic_df_1.drop("Survived", axis=1)
        titanic_df_2 = titanic_df_2.drop("Survived", axis=1)
        varcat = ["Pclass", "Sex", "Embarked", "Title"]
        y = titanic_df["Survived"]
        X = titanic_df.drop("Survived", axis=1)
        categ_encoding = OrdinalEncoder(cols=varcat, handle_unknown="ignore", return_df=True).fit(X)
        X = categ_encoding.transform(X)
        x_train, x_test, y_train, y_test = train_test_split(X, y, train_size=0.75, random_state=1)
        rf = RandomForestClassifier(n_estimators=100, min_samples_leaf=3)
        rf.fit(x_train, y_train)
        self.titanic_df_1 = titanic_df_1
        self.titanic_df_2 = titanic_df_2
        self.rf = rf
        self.categ_encoding = categ_encoding
        self.script_path = script_path
        self.X = X

    def test_init_1(self):
        """
        test init 1 SmartDrift
        """
        smart_drift = SmartDrift()
        assert hasattr(smart_drift, "df_baseline")

    def test_compile_nooptions(self):
        """
        Test compile()
        """
        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        smart_drift.compile()
        assert isinstance(smart_drift.xpl, shapash.explainer.smart_explainer.SmartExplainer)

    def test_compile_fullvalid(self):
        """
        test compile() with full validation
        """
        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        smart_drift.compile(full_validation=True)
        assert isinstance(smart_drift.xpl, shapash.explainer.smart_explainer.SmartExplainer)

    def test_compile_nosampling(self):
        """
        test compile() without no sampling
        """
        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        smart_drift.compile(full_validation=True, sampling=False)
        assert isinstance(smart_drift.xpl, shapash.explainer.smart_explainer.SmartExplainer)

    def test_compile_samplingsize(self):
        """
        test compile() with a fixed sample size
        """
        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        smart_drift.compile(full_validation=True, sample_size=100)
        assert isinstance(smart_drift.xpl, shapash.explainer.smart_explainer.SmartExplainer)

    def test_compile_model_encoder(self):
        """
        test compile() with a model and an encoder specified
        """
        smart_drift = SmartDrift(
            self.titanic_df_1, self.titanic_df_2, deployed_model=self.rf, encoding=self.categ_encoding
        )
        smart_drift.compile()
        assert isinstance(smart_drift.xpl, shapash.explainer.smart_explainer.SmartExplainer)

    def test_compile_dataset_names(self):
        """
        test compile() with a model and an encoder specified
        """
        smart_drift = SmartDrift(
            self.titanic_df_1,
            self.titanic_df_2,
            deployed_model=self.rf,
            encoding=self.categ_encoding,
            dataset_names={"df_current": "titanic 2", "df_baseline": "titanic 1"},
        )
        smart_drift.compile()
        assert isinstance(smart_drift.xpl, shapash.explainer.smart_explainer.SmartExplainer)

    def test_generate_report_fullvalid(self):
        """
        test generate_report() with fullvalidation option specified to True
        """
        smart_drift = SmartDrift(
            self.titanic_df_1, self.titanic_df_2, deployed_model=self.rf, encoding=self.categ_encoding
        )
        smart_drift.compile(full_validation=True)
        infofile = path.join(self.script_path, PROJECT_INFO_PATH)
        smart_drift.generate_report(output_file=REPORT_PATH, project_info_file=infofile)

        assert path.isfile(REPORT_PATH)
        os.remove(REPORT_PATH)

    def test_generate_report_nofullvalid(self):
        """
        test generate_report() with no full validation
        """
        smart_drift = SmartDrift(
            self.titanic_df_1, self.titanic_df_2, deployed_model=self.rf, encoding=self.categ_encoding
        )
        smart_drift.compile()
        infofile = path.join(self.script_path, PROJECT_INFO_PATH)
        smart_drift.generate_report(output_file=REPORT_PATH, project_info_file=infofile)
        assert path.isfile(REPORT_PATH)
        os.remove(REPORT_PATH)

    def test_add_data_modeldrift_1(self):
        """
        test add_data_modeldrift method
        """
        lift_info = {
            "annee": ["2019", "2019"],
            "mois": ["7", "8"],
            "plage_historique": [1, 1],
            "target": ["souscription", "souscription"],
            "type_lift": ["lift10", "lift10"],
            "performance": [3.32, 3.18],
        }
        df_lift = pd.DataFrame(lift_info)
        smart_drift = SmartDrift(
            self.titanic_df_1, self.titanic_df_2, deployed_model=self.rf, encoding=self.categ_encoding
        )
        smart_drift.compile(ignore_cols=["PassengerId"])
        assert isinstance(smart_drift.data_modeldrift, type(None))
        smart_drift.add_data_modeldrift(dataset=df_lift)
        assert isinstance(smart_drift.data_modeldrift, pd.core.frame.DataFrame)

    def test_add_data_modeldrift_2(self):
        """
        test add_data_modeldrift method
        """
        lift_info = {
            "annee": ["2019", "2019"],
            "mois": ["7", "8"],
            "plage_historique": [1, 1],
            "target": ["souscription", "souscription"],
            "type_lift": ["lift10", "lift10"],
            "lift": [3.32, 3.18],
        }
        df_lift = pd.DataFrame(lift_info)
        smart_drift = SmartDrift(
            self.titanic_df_1, self.titanic_df_2, deployed_model=self.rf, encoding=self.categ_encoding
        )
        smart_drift.compile(ignore_cols=["PassengerId"])
        assert isinstance(smart_drift.data_modeldrift, type(None))
        with self.assertRaises(Exception):
            smart_drift.add_data_modeldrift(dataset=df_lift)
        smart_drift.add_data_modeldrift(dataset=df_lift, metric="lift")
        assert isinstance(smart_drift.data_modeldrift, pd.core.frame.DataFrame)
        smart_drift.add_data_modeldrift(
            dataset=df_lift, metric="lift", reference_columns=["target", "plage_historique"]
        )
        assert isinstance(smart_drift.data_modeldrift, pd.core.frame.DataFrame)

    def test_analyse_consistency_nofullvalidation_1(self):
        """
        Test _analyze_consistency() method with full validation defined to False
        """
        script_path = Path(path.abspath(__file__)).parent.parent.parent.parent
        titanic_original = path.join(script_path, TITANIC_ORIGINAL_PATH)
        df_original = pd.read_csv(titanic_original, index_col=0)
        titanic_altered = path.join(script_path, TITANIC_ALTERED_PATH)
        df_altered = pd.read_csv(titanic_altered, index_col=0)

        df_altered["col_sup"] = 0
        smart_drift = SmartDrift(df_altered, df_original)
        pb_cols, err_mods = smart_drift._analyze_consistency(
            full_validation=False, ignore_cols=["col_sup", "Name", "PassengerId"]
        )

        expected_pb_cols = {"New columns": [], "Removed columns": ["col_sup"], "Type errors": []}

        assert isinstance(pb_cols, dict)
        assert all(key in ("New columns", "Removed columns", "Type errors") for key in pb_cols.keys())
        assert isinstance(err_mods, dict)
        assert pb_cols == expected_pb_cols
        assert err_mods == dict()

    def test_analyse_consistency_nofullvalidation_2(self):
        """
        Test _analyze_consistency() method with full validation defined to False
        """
        script_path = Path(path.abspath(__file__)).parent.parent.parent.parent
        titanic_original = path.join(script_path, TITANIC_ORIGINAL_PATH)
        df_original = pd.read_csv(titanic_original, index_col=0)
        titanic_altered = path.join(script_path, TITANIC_ALTERED_PATH)
        df_altered = pd.read_csv(titanic_altered, index_col=0)

        df_altered["col_sup"] = 0
        smart_drift = SmartDrift(df_altered, df_original)
        pb_cols, err_mods = smart_drift._analyze_consistency(
            full_validation=False, ignore_cols=["col_sup", "Name", "PassengerId"]
        )

        with self.assertRaises(TypeError):
            smart_drift.df_current = dict()
            smart_drift._analyze_consistency(full_validation=False, ignore_cols=["col_sup", "Name", "PassengerId"])

    def test_analyse_consistency_fullvalidation(self):
        """
        Test _analyze_consistency() method with full validation defined to True
        """
        script_path = Path(path.abspath(__file__)).parent.parent.parent.parent
        titanic_original = path.join(script_path, TITANIC_ORIGINAL_PATH)
        df_original = pd.read_csv(titanic_original, index_col=0)
        titanic_altered = path.join(script_path, TITANIC_ALTERED_PATH)
        df_altered = pd.read_csv(titanic_altered, index_col=0)

        df_altered["col_sup"] = 0
        df_altered = df_altered.replace("male", "Male")
        df_original["col_sup"] = "0"
        smart_drift = SmartDrift(df_altered, df_original)
        pb_cols, err_mods = smart_drift._analyze_consistency(
            full_validation=True, ignore_cols=["Title", "Name", "PassengerId"]
        )

        expected_err_mods = {"Sex": {"New distinct values": ["Male"], "Removed distinct values": ["male"]}}

        expected_pb_cols = {"New columns": [], "Removed columns": [], "Type errors": ["col_sup"]}

        assert isinstance(pb_cols, dict)
        assert all(key in ("New columns", "Removed columns", "Type errors") for key in pb_cols.keys())
        assert isinstance(err_mods, dict)
        assert pb_cols == expected_pb_cols
        assert err_mods == expected_err_mods

    def test_predict_1(self):
        """
        test _predict() method
        """

        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        df_predict = smart_drift._predict()
        assert df_predict is None

        with self.assertRaises(Exception):
            smart_drift._predict(deployed_model=self.rf)

        df_predict = smart_drift._predict(deployed_model=self.rf, encoding=self.categ_encoding)
        assert isinstance(df_predict, pd.DataFrame)
        assert all(column in ("dataset", "Score") for column in df_predict.columns)
        assert df_predict.shape[0] == (smart_drift.df_baseline.shape[0] + smart_drift.df_current.shape[0])
        assert all(col_type in ("float64", "object") for col_type in df_predict.dtypes)

    def test_predict_2(self):
        """
        test _predict() method
        """

        varcat = ["Pclass", "Sex", "Embarked", "Title"]

        false_categ_encoding = OrdinalEncoder(cols=varcat, handle_unknown="ignore", return_df=True).fit(
            self.X[["Title", "Pclass", "Sex", "Embarked"]]
        )
        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)

        with self.assertRaises(Exception):
            smart_drift._predict(deployed_model=self.rf, encoding=false_categ_encoding)

        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        df_predict = smart_drift._predict(deployed_model=self.rf, encoding=self.categ_encoding)

        assert isinstance(df_predict, pd.DataFrame)
        assert all(column in ("dataset", "Score") for column in df_predict.columns)
        assert df_predict.shape[0] == (smart_drift.df_baseline.shape[0] + smart_drift.df_current.shape[0])
        assert all(col_type in ("float64", "object") for col_type in df_predict.dtypes)

    def test_predict_3(self):
        """
        test _predict() method
        """
        varcat = ["Pclass", "Sex", "Embarked", "Title"]
        false_categ_encoding = OrdinalEncoder(cols=varcat, handle_unknown="ignore", return_df=True).fit(
            self.X[["Title", "Pclass", "Sex", "Embarked"]]
        )
        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)

        with self.assertRaises(Exception):
            smart_drift._predict(deployed_model=self.rf, encoding=false_categ_encoding)

        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)

        with self.assertRaises(Exception):
            smart_drift._predict(deployed_model=self.rf, encoding=false_categ_encoding)

        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        df_predict = smart_drift._predict(deployed_model=self.rf, encoding=self.categ_encoding)

        assert isinstance(df_predict, pd.DataFrame)
        assert all(column in ("dataset", "Score") for column in df_predict.columns)
        assert df_predict.shape[0] == (smart_drift.df_baseline.shape[0] + smart_drift.df_current.shape[0])
        assert all(col_type in ("float64", "object") for col_type in df_predict.dtypes)

        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        df_predict = smart_drift._predict(deployed_model=self.rf, encoding=self.categ_encoding)
        assert isinstance(df_predict, pd.DataFrame)
        assert all(column in ("dataset", "Score") for column in df_predict.columns)
        assert df_predict.shape[0] == (smart_drift.df_baseline.shape[0] + smart_drift.df_current.shape[0])
        assert all(col_type in ("float64", "object") for col_type in df_predict.dtypes)

    def test_feature_importance(self):
        """
        Test on _feature_importance()'s method
        """
        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        smart_drift.compile()

        feature_imp = smart_drift._feature_importance()
        assert feature_imp is None

        feature_imp = smart_drift._feature_importance(deployed_model=self.rf)
        assert isinstance(feature_imp, pd.DataFrame)
        assert all(column in ("feature", "deployed_model", "datadrift_classifier") for column in feature_imp.columns)
        assert all(col_type in ("float64", "object") for col_type in feature_imp.dtypes)

    def test_histo_datadrift_metric_1(self):
        """
        Test on _histo_datadrift_metric() method
        """
        SD = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        df_auc = SD._histo_datadrift_metric()
        assert df_auc == None

        SD = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        SD.compile(datadrift_file="tests/data/AUC_Histo.csv")
        df_auc = SD._histo_datadrift_metric()
        assert isinstance(df_auc, pd.DataFrame)
        assert all(column in ("date", "auc") for column in df_auc.columns)
        assert all(column in ("object", "float64") for column in df_auc.dtypes)
        os.remove("tests/data/AUC_Histo.csv")

    def test_histo_datadrift_metric_2(self):
        """
        Test on _histo_datadrift_metric() method
        """
        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        smart_drift.compile(datadrift_file="tests/data/AUC_Histo.csv")

        with self.assertRaises(Exception):
            smart_drift._histo_datadrift_metric(
                datadrift_file="tests/data/AUC_Histo.csv", predict_test=smart_drift.model.predict(self.titanic_df_2)
            )

        df_auc = smart_drift._histo_datadrift_metric()
        assert isinstance(df_auc, pd.DataFrame)
        assert all(column in ("date", "auc") for column in df_auc.columns)
        assert all(column in ("object", "float64") for column in df_auc.dtypes)
        os.remove("tests/data/AUC_Histo.csv")

    def test_histo_datadrift_metric_3(self):
        """
        Test on _histo_datadrift_metric() method
        """
        smart_drift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        smart_drift.compile(datadrift_file="tests/data/AUC_Histo.csv")

        with self.assertRaises(Exception):
            smart_drift._histo_datadrift_metric(date_compile_auc="01-12-2020")

        datadrift_file = path.join(self.script_path, "tests/data/AUC_Histo.csv")
        df_auc = smart_drift._histo_datadrift_metric(date_compile_auc="01/12/2020", datadrift_file=datadrift_file)
        assert isinstance(df_auc, pd.DataFrame)
        assert all(column in ("date", "auc") for column in df_auc.columns)
        assert all(column in ("object", "float64") for column in df_auc.dtypes)
        os.remove("tests/data/AUC_Histo.csv")

    def test_histo_datadrift_metric_4(self):
        """
        Test on _histo_datadrift_metric() method
        """
        smart_drift = SmartDrift(
            self.titanic_df_1, self.titanic_df_2, deployed_model=self.rf, encoding=self.categ_encoding
        )
        smart_drift.compile(datadrift_file="tests/data/AUC_Histo.csv")

        with self.assertRaises(Exception):
            smart_drift._histo_datadrift_metric(date_compile_auc="01-12-2020")

        datadrift_file = path.join(self.script_path, "tests/data/AUC_Histo.csv")
        df_auc = smart_drift._histo_datadrift_metric(date_compile_auc="01/12/2020", datadrift_file=datadrift_file)
        assert isinstance(df_auc, pd.DataFrame)
        assert all(column in ("date", "auc", "JS_predict") for column in df_auc.columns)
        assert all(column in ("object", "float64") for column in df_auc.dtypes)
        os.remove("tests/data/AUC_Histo.csv")

    def test_save_load(self):
        """
        Test on save() and load() methods : checks that the attributes meant to be saved and loaded are saved and loaded
        """
        temp_pkl_path = "tests/data/test_pkl.pkl"
        sd = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        sd.compile(datadrift_file="tests/data/AUC_Histo.csv")
        sd.save(temp_pkl_path)
        sd2 = SmartDrift.load(temp_pkl_path)
        assert isinstance(sd2, SmartDrift)
        pd.testing.assert_frame_equal(sd.df_baseline, sd2.df_baseline)
        assert len(sd.df_baseline) == len(sd2.df_baseline)
        self.assertCountEqual(sd.df_current.columns, sd2.df_current.columns)
        assert len(sd.df_current) == len(sd2.df_current)
        assert sd.feature_importance == sd2.feature_importance
        self.assertCountEqual(sd.ignore_cols, sd2.ignore_cols)
        self.assertCountEqual(sd.pb_cols, sd2.pb_cols)
        self.assertCountEqual(sd.err_mods, sd2.err_mods)
        assert sd.auc == sd2.auc
        pd.testing.assert_frame_equal(sd.historical_auc, sd2.historical_auc)
        pd.testing.assert_frame_equal(sd._df_concat, sd2._df_concat)
        assert sd._datadrift_target == sd2._datadrift_target
        assert sd.deployed_model == sd2.deployed_model
        assert sd.encoding == sd2.encoding
        assert sd.palette_name == sd2.palette_name
        assert sd.colors_dict == sd2.colors_dict
        os.remove(temp_pkl_path)

    def test_define_style(self):
        """
        test define_style() method : checks that the frontend style defined in colors.json is written into the SmartDrift plot attribute
        """
        sd = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        sd.compile(datadrift_file="tests/data/AUC_Histo.csv")
        colors_dict = select_palette(colors_loading(), "eurybia")
        with self.assertRaises(ValueError):
            sd.define_style()
        sd.define_style(palette_name="eurybia", colors_dict=colors_dict)
        assert sd.colors_dict == colors_dict
        assert sd.plot._style_dict["univariate_cat_bar"] == colors_dict["univariate_cat_bar"]
        assert sd.plot._style_dict["univariate_cont_bar"] == colors_dict["univariate_cont_bar"]
        assert sd.plot._style_dict["datadrift_historical"] == colors_dict["datadrift_historical"]
        assert sd.plot._style_dict["scatter_plot"] == colors_dict["scatter_plot"]
        assert sd.plot._style_dict["scatter_line"] == colors_dict["scatter_line"]
        assert sd.plot._style_dict["featimportance_colorscale"] == colors_dict["featimportance_colorscale"]
        assert sd.plot._style_dict["contrib_colorscale"] == colors_dict["contrib_colorscale"]
        # not testing the shapash.explainer.smart_explainer

    def test_datetime_column_transformation(self):
        """
        Test if SmartDrift can automatically handle datatime columns
        """

        date_list = pd.date_range(start="01/01/2022", end="01/30/2022")
        X1 = np.random.rand(len(date_list))
        X2 = np.random.rand(len(date_list))

        df_current = pd.DataFrame(date_list, columns=["date"])
        df_current["col1"] = X1
        df_baseline = pd.DataFrame(date_list, columns=["date"])
        df_baseline["col1"] = X2

        sd = SmartDrift(df_current=df_current, df_baseline=df_baseline)
        sd.compile(full_validation=True)
        # Should pass this step
        auc = sd.auc
        assert auc > 0

    def test_datetime_column_model_error(self):
        """
        Test if SmartDrift raised an error when their is datatime columns
        and deployed_model is filled
        """

        date_list = pd.date_range(start="01/01/2022", end="01/30/2022")
        X1 = np.random.rand(len(date_list))
        X2 = np.random.rand(len(date_list))

        df_current = pd.DataFrame(date_list, columns=["date"])
        df_current["col1"] = X1
        df_baseline = pd.DataFrame(date_list, columns=["date"])
        df_baseline["col1"] = X2

        # Random models
        regressor = RandomForestRegressor(n_estimators=2).fit(df_baseline[["col1"]], df_baseline["col1"].ravel())

        sd = SmartDrift(df_current=df_current, df_baseline=df_baseline, deployed_model=regressor)

        # Should raise an error
        with pytest.raises(TypeError, match="df_current have datetime column. You should drop it"):
            sd.compile(full_validation=True)



"""
Integration test for generation.py
"""

import os
import unittest
from os import path
from pathlib import Path

import pandas as pd
from category_encoders import OrdinalEncoder
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split

from eurybia import SmartDrift
from eurybia.report.generation import execute_report

current_path = os.path.dirname(os.path.abspath(__file__))


class TestGeneration(unittest.TestCase):
    """
    Unit test generation.py
    """

    def setUp(self) -> None:
        """
        Initialize data for testing part
        """
        script_path = Path(path.abspath(__file__)).parent.parent.parent
        titanic_original = path.join(script_path, "eurybia/data/titanicdata.csv")
        titan_df = pd.read_csv(titanic_original, index_col=0)
        y = titan_df["Survived"]
        X = titan_df.drop("Survived", axis=1).drop("Name", axis=1)
        varcat = ["Pclass", "Sex", "Embarked", "Title"]
        categ_encoding = OrdinalEncoder(cols=varcat, handle_unknown="ignore", return_df=True).fit(X)
        X_encoded = categ_encoding.transform(X)

        Xtrain, Xtest, ytrain, ytest = train_test_split(X, y, train_size=0.75, random_state=1)
        Xtrain_e, Xtest_e, ytrain_e, ytest_e = train_test_split(X_encoded, y, train_size=0.75, random_state=1)

        rf = RandomForestClassifier(n_estimators=100, min_samples_leaf=3)
        rf.fit(Xtrain_e, ytrain_e)

        smartdrift = SmartDrift(Xtrain, Xtest, deployed_model=rf, encoding=categ_encoding)
        smartdrift.compile(full_validation=True)
        self.smartdrift = smartdrift
        self.xpl = smartdrift.xpl
        self.script_path = script_path

    def tearDown(self) -> None:
        """
        method that tidies up after the test method has been run
        """
        os.remove("./report.html")

    def test_execute_report_1(self) -> None:
        """
        Test execute_report() method
        """
        execute_report(
            smartdrift=self.smartdrift,
            explainer=self.xpl,
            project_info_file=os.path.join(current_path, "../data/project_info.yml"),
            output_file="./report.html",
        )

        assert os.path.exists("./report.html")

    def test_execute_report_2(self) -> None:
        """
        Test execute_report() method
        """

        execute_report(
            smartdrift=self.smartdrift,
            explainer=self.xpl,
            project_info_file=os.path.join(current_path, "../data/project_info.yml"),
            output_file="./report.html",
        )

        assert os.path.exists("./report.html")

    def test_execute_report_3(self) -> None:
        """
        Test execute_report() method
        """
        execute_report(
            smartdrift=self.smartdrift,
            explainer=self.xpl,
            project_info_file=os.path.join(current_path, "../data/project_info.yml"),
            output_file="./report.html",
            config_report=dict(title_story="Test integration", title_description="Title of test integration"),
        )

        assert os.path.exists("./report.html")

    def test_execute_report_modeldrift_1(self) -> None:
        """
        Test execute_report() method
        """
        df_perf = pd.DataFrame({"mois": [1, 2, 3], "annee": [2018, 2019, 2020], "performance": [2, 3.46, 2.5]})
        self.smartdrift.add_data_modeldrift(dataset=df_perf, metric="performance")
        execute_report(
            smartdrift=self.smartdrift,
            explainer=self.xpl,
            project_info_file=os.path.join(current_path, "../data/project_info.yml"),
            output_file="./report.html",
            config_report=dict(title_story="Test integration", title_description="Title of test integration"),
        )

        assert os.path.exists("./report.html")

    def test_execute_report_modeldrift_2(self) -> None:
        """
        Test execute_report() method
        """

        import random

        annees = [2020, 2020, 2021, 2021, 2021] * 6
        mois = [11, 12, 1, 2, 3] * 6
        historical_range = [3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1] * 2
        metric = ["lift_devis"] * 15 + ["lift_souscription"] * 15
        metric_value = [random.uniform(3.0, 7.0) for _ in range(30)]

        df_perf2 = pd.DataFrame(
            {
                "annee": annees,
                "mois": mois,
                "historical_range": historical_range,
                "metric": metric,
                "metric_value": metric_value,
            }
        )
        self.smartdrift.add_data_modeldrift(
            dataset=df_perf2, metric="metric_value", reference_columns=["historical_range", "metric"]
        )
        script_path = Path(path.abspath(__file__)).parent.parent.parent
        titanic_original = path.join(script_path, "eurybia/data/project_info_titanic.yml")
        execute_report(
            smartdrift=self.smartdrift,
            explainer=self.xpl,
            project_info_file=titanic_original,
            output_file="./report.html",
            config_report=dict(title_story="Test integration", title_description="Title of test integration"),
        )

        assert os.path.exists("./report.html")



"""
Integration test for main object Smartdrift
"""

import os
import unittest
from os import path
from pathlib import Path

import pandas as pd
import plotly
from category_encoders import OrdinalEncoder
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split

from eurybia import SmartDrift

current_path = os.path.dirname(os.path.abspath(__file__))
TITANIC_PATH = "eurybia/data/titanicdata.csv"
TITANIC_ORIGINAL_PATH = "eurybia/data/titanic_original.csv"
TITANIC_ALTERED_PATH = "eurybia/data/titanic_altered.csv"
PROJECT_INFO_PATH = "eurybia/data/project_info_titanic.yml"
REPORT_PATH = "./test_report.html"


class TestIntegrationSmartdrift(unittest.TestCase):
    """
    Integration test smartdrift
    """

    def setUp(self):
        """
        Initialize data for testing part
        """
        script_path = Path(path.abspath(__file__)).parent.parent.parent
        titanic_csv = path.join(script_path, TITANIC_PATH)
        titanic_df = pd.read_csv(titanic_csv)
        del titanic_df["Name"]
        titanic_df = titanic_df.drop("PassengerId", axis=1)
        titanic_df_1, titanic_df_2 = train_test_split(titanic_df, test_size=0.5, random_state=42)
        titanic_df_1 = titanic_df_1.drop("Survived", axis=1)
        titanic_df_2 = titanic_df_2.drop("Survived", axis=1)
        varcat = ["Pclass", "Sex", "Embarked", "Title"]
        y = titanic_df["Survived"]
        X = titanic_df.drop("Survived", axis=1)
        categ_encoding = OrdinalEncoder(cols=varcat, handle_unknown="ignore", return_df=True).fit(X)
        X = categ_encoding.transform(X)
        x_train, x_test, y_train, y_test = train_test_split(X, y, train_size=0.75, random_state=1)
        rf = RandomForestClassifier(n_estimators=100, min_samples_leaf=3)
        rf.fit(x_train, y_train)
        self.titanic_df_1 = titanic_df_1
        self.titanic_df_2 = titanic_df_2
        self.rf = rf
        self.categ_encoding = categ_encoding
        self.script_path = script_path
        self.X = X

    def test_compile_smartdrift_1(self):
        """
        Test compile smartdrift
        """

        smartdrift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        smartdrift.compile(full_validation=False)
        fig_continuous = smartdrift.plot.generate_fig_univariate("Age")
        fig_categorical = smartdrift.plot.generate_fig_univariate("Pclass")

        assert smartdrift.auc < 0.6

        assert isinstance(fig_continuous, plotly.graph_objs._figure.Figure)
        assert isinstance(fig_categorical, plotly.graph_objs._figure.Figure)

    def test_compile_smartdrift_2(self):
        """
        Test compile smartdrift
        """

        smartdrift = SmartDrift(
            self.titanic_df_1, self.titanic_df_2, deployed_model=self.rf, encoding=self.categ_encoding
        )
        smartdrift.compile(full_validation=True)
        fig_continuous = smartdrift.plot.generate_fig_univariate("Age")
        fig_categorical = smartdrift.plot.generate_fig_univariate("Pclass")
        fig_predict = smartdrift.plot.generate_fig_univariate(df_all=smartdrift.df_predict, col="Score", hue="dataset")

        assert smartdrift.auc < 0.6
        assert (
            smartdrift.feature_importance.loc[smartdrift.feature_importance["feature"] == "Age", "deployed_model"].iloc[
                0
            ]
            > 0.1
        )
        assert (
            smartdrift.feature_importance.loc[
                smartdrift.feature_importance["feature"] == "Age", "datadrift_classifier"
            ].iloc[0]
            > 0.1
        )

        assert isinstance(fig_continuous, plotly.graph_objs._figure.Figure)
        assert isinstance(fig_categorical, plotly.graph_objs._figure.Figure)
        assert isinstance(fig_predict, plotly.graph_objs._figure.Figure)

    def test_compile_smartdrift_3(self):
        """
        Test compile smartdrift
        """

        smartdrift = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        smartdrift.compile(date_compile_auc="01/01/2021", datadrift_file="datadrift_metric.csv")

        smartdrift2 = SmartDrift(self.titanic_df_1, self.titanic_df_2)
        smartdrift2.compile(date_compile_auc="01/01/2022", datadrift_file="datadrift_metric.csv")

        fig_continuous = smartdrift2.plot.generate_fig_univariate("Age")
        fig_categorical = smartdrift2.plot.generate_fig_univariate("Pclass")

        assert smartdrift2.auc < 0.6

        assert isinstance(fig_continuous, plotly.graph_objs._figure.Figure)
        assert isinstance(fig_categorical, plotly.graph_objs._figure.Figure)
        os.remove("datadrift_metric.csv")



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import os
import subprocess
import sys

sys.path.insert(0, "../..")

import eurybia  # E402

# -- Project information -----------------------------------------------------

project = "Eurybia"
copyright = "2022, Maif"
author = "Maif"

# The short X.Y version
version = eurybia.__version__

# The full version, including alpha/beta/rc tags
release = eurybia.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.todo",
    "sphinx.ext.napoleon",
    "nbsphinx",
]
nbsphinx_execute = "never"
master_doc = "index"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["../_templates"]

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_material"
# html_logo = "../_static/eurybia-fond-clair.png"

html_theme_options = {
    # Set the name of the project to appear in the navigation.
    "nav_title": "Eurybia",
    # Set the color and the accent color
    "color_primary": "light-blue",
    "color_accent": "blue",
    # Set the repo location to get a badge with stats
    "repo_url": "https://github.com/MAIF/eurybia",
    "repo_name": "eurybia",
    # Icon of the navbar
    "logo_icon": "&#xe873",
    # Visible levels of the global TOC; -1 means unlimited
    "globaltoc_depth": 3,
    # If False, expand all TOC entries
    "globaltoc_collapse": True,
    # If True, show hidden TOC entries
    "globaltoc_includehidden": False,
}

html_sidebars = {"**": ["logo-text.html", "globaltoc.html", "localtoc.html", "searchbox.html"]}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["../_static"]


# -- Extension configuration -------------------------------------------------

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# -- Additional html pages  -------------------------------------------------
# Generates the report example in the documentation
subprocess.run(["python", "../../tutorial/Eurybia_report_example.py"])
html_extra_path = ["../../tutorial/report.html"]


def setup_tutorials():
    """
    copy all notebooks tutorials in docs to generate rst tutorials
    """
    import pathlib
    import shutil

    def _copy_notebooks_and_create_rst(d_path, new_d_path, d_name):
        os.makedirs(new_d_path, exist_ok=True)
        list_notebooks = [f for f in os.listdir(d_path) if os.path.splitext(f)[-1] == ".ipynb"]
        for notebook_f_name in list_notebooks:
            shutil.copyfile(os.path.join(d_path, notebook_f_name), os.path.join(new_d_path, notebook_f_name))

        # RST file (see for example docs/overview.rst)
        rst_file = "\n".join(
            [
                f"{d_name}",
                "======================",
                "",
                ".. toctree::",
                "    :maxdepth: 1",
                "    :glob:",
                "",
                "    *",
                "",
            ]
        )
        with open(os.path.join(new_d_path, "index.rst"), "w") as f:
            f.write(rst_file)

    docs_path = pathlib.Path(__file__).parent
    tutorials_path = os.path.join(docs_path.parent.parent, "tutorial")
    tutorials_doc_path = os.path.join(docs_path, "tutorials")

    # Create a directory in eurybia/docs/source/tutorials for each directory of eurybia/tutorial
    # And copy each notebook file in it
    list_dir = [
        d
        for d in os.listdir(tutorials_path)
        if os.path.isdir(os.path.join(tutorials_path, d)) and not d.startswith(".")
    ]
    for d_name in list_dir:
        d_path = os.path.join(tutorials_path, d_name)
        new_d_path = os.path.join(tutorials_doc_path, d_name)
        _copy_notebooks_and_create_rst(d_path, new_d_path, d_name)

    # Also copying all the overview tutorials
    _copy_notebooks_and_create_rst(tutorials_path, tutorials_doc_path, "overview")


setup_tutorials()



"""
Unit test statistical_tests.py
"""
import unittest

import numpy as np

from eurybia.utils.statistical_tests import chisq_test, ksmirnov_test, compute_js_divergence, prob_mass_fun


class TestStatistical_tests(unittest.TestCase):
    """
    Unit test statistical_tests.py
    """

    def test_ksmirnov_test(self):
        """
        Unit test ksmirnov_test function
        """
        X1 = np.array([0, 3, 9857, 3444, 99, 8, 9, 9, 0, 4535, 34])
        X2 = np.array([0.45, 3, 9857, 3444, 4535, 34])

        res = ksmirnov_test(X1, X2)

        assert res["testname"] == "K-Smirnov"
        assert round(res["statistic"], 2) == 0.23
        assert round(res["pvalue"], 2) == 0.95

    def test_chisq_test(self):
        """
        Unit test chisq_test function
        """
        X1 = np.array([0, 0, 3, 4, 2, 2, 2, 2, 2, 1])
        X2 = np.array([3, 4, 3, 2, 1])

        res = chisq_test(X1, X2)

        assert res["testname"] == "Chi-Square"
        assert round(res["statistic"], 2) == 3.75
        assert round(res["pvalue"], 2) == 0.44

    def test_prob_mass_fun(self):
        """
        Unit test prob_mass_fun
        """
        X1 = np.array([0, 0, 3, 3, 0, 2, 2, 2, 1, 1])
        X2 = np.array([3, 3, 3, 2, 1])

        a = np.concatenate((X1, X2), axis=0)
        range = (a.min(), a.max())

        res1, res2 = prob_mass_fun(X1,3,range)

        assert res1.all() == np.array([0., 1., 2., 3.]).all()
        assert res2.all() == np.array([0.3, 0.2 , 0.5]).all()

    def test_compute_js_divergence(self):
        """
        Unit test Jensen Shannon divergence
        """
        X1 = np.array([0, 3, 9857, 3444, 99, 8, 9, 9, 0, 4535, 34])
        X2 = np.array([0.45, 3, 9857, 3444, 4535, 34])

        X3 = np.array([0, 0, 3, 4, 2, 2, 2, 2, 2, 1])
        X4 = np.array([3, 4, 3, 2, 1])

        res1 = compute_js_divergence(X1, X2)
        res2 = compute_js_divergence(X3, X4)

        assert round(res1, 2) == 0.17
        assert round(res2, 2) == 0.41



"""Top-level package."""

__author__ = """Thomas Bouche, Johann Martin, Nicolas Roux"""
__email__ = "thomas.bouche@maif.fr"

from eurybia.core.smartdrift import SmartDrift

VERSION = (1, 2, 0)

__version__ = ".".join(map(str, VERSION))

__all__ = ["SmartDrift"]



"""
IO module
"""

import pickle


def load_yml(path):
    """
    Loads a yml file
    Parameters
    ----------
    path : str
        File path where the yml file is stored.
    Returns
    -------
    dict_yaml : dict
        Python dict containing the parsed yml file.
    """
    if not isinstance(path, str):
        raise ValueError(
            """
            path parameter must be a string
            """
        )

    with open(path) as file:
        dict_yaml = yaml.full_load(file)

    return dict_yaml


try:
    import yaml

    _is_yaml_available = True
except (ImportError, ModuleNotFoundError):
    _is_yaml_available = False


def save_pickle(obj, path, protocol=pickle.HIGHEST_PROTOCOL):
    """
    Save any python Object in pickle file
    Parameters
    ----------
    obj : any Python Object
    path : str
        File path where the pickled object will be stored.
    protocol : int
        Int which indicates which protocol should be used by the pickler,
        default HIGHEST_PROTOCOL
    """

    if not isinstance(path, str):
        raise ValueError(
            """
            path parameter must be a string
            """
        )

    if not isinstance(protocol, int):
        raise ValueError(
            """
            protocol parameter must be an integer
            """
        )
    with open(path, "wb") as file:
        pickle.dump(obj, file, protocol=protocol)


def load_pickle(path):
    """
    load any pickle file
    Parameters
    ----------
    path : str
        File path where the pickled object is stored.
    Returns
    -------
    object that pickle file contains
    """

    if not isinstance(path, str):
        raise ValueError(
            """
            path parameter must be a string
            """
        )

    with open(path, "rb") as file:
        pklobj = pickle.load(file)

    return pklobj



"""
Unit test io.py
"""
import unittest
from os import path
from pathlib import Path

from eurybia.utils.io import load_yml


class Testio(unittest.TestCase):
    """
    Unit test io.py
    """

    def test_load_yml(self):
        """
        Unit test load_yml method
        """
        script_path = Path(path.abspath(__file__)).parent.parent.parent
        project_info_path = path.join(script_path, "data/project_info.yml")

        with self.assertRaises(ValueError):
            load_yml(path=0)

        load_yml(project_info_path)



"""
Data loader module
"""

import json
import os
from urllib.request import urlretrieve

import pandas as pd


def data_loading(dataset):
    """
    data_loading allows Eurybia user to try the library with small but clear datasets.
    Titanic, house_prices or us_car_accident data.

    Example
    ----------
    >>> from eurybia.data.data_loader import data_loading
    >>> house_df, house_dict = data_loading('house_prices')

    Parameters
    ----------
    dataset : String
        Dataset's name to return.
         - 'titanic'
         - 'house_prices'
         - 'us_car_accident'

    Returns
    -------
    data : pandas.DataFrame
        Dataset required
    dict : (Dictionnary, Optional)
        If exist, columns labels dictionnary associated to the dataset.
    """
    current_path = os.path.dirname(os.path.abspath(__file__))
    if dataset == "house_prices":
        if os.path.isfile(current_path + "/house_prices_dataset.csv") is False:
            github_data_url = "https://github.com/MAIF/eurybia/raw/master/eurybia/data/"
            urlretrieve(
                github_data_url + "house_prices_dataset.csv", filename=current_path + "/house_prices_dataset.csv"
            )
            urlretrieve(
                github_data_url + "house_prices_labels.json", filename=current_path + "/house_prices_labels.json"
            )
        data_house_prices_path = os.path.join(current_path, "house_prices_dataset.csv")
        dict_house_prices_path = os.path.join(current_path, "house_prices_labels.json")
        data = pd.read_csv(data_house_prices_path, header=0, index_col=0, engine="python")
        with open(dict_house_prices_path) as openfile2:
            dic = json.load(openfile2)
        return data, dic

    elif dataset == "titanic":
        if os.path.isfile(current_path + "/titanicdata.csv") is False:
            github_data_url = "https://github.com/MAIF/eurybia/raw/master/eurybia/data/"
            urlretrieve(github_data_url + "titanicdata.csv", filename=current_path + "/titanicdata.csv")
        data_titanic_path = os.path.join(current_path, "titanicdata.csv")
        data = pd.read_csv(data_titanic_path, header=0, index_col=0, engine="python")
        return data

    elif dataset == "us_car_accident":
        if os.path.isfile(current_path + "/US_Accidents_extract.csv") is False:
            github_data_url = "https://github.com/MAIF/eurybia/raw/master/eurybia/data/"
            urlretrieve(
                github_data_url + "US_Accidents_extract.csv", filename=current_path + "/US_Accidents_extract.csv"
            )
        data_us_car_path = os.path.join(current_path, "US_Accidents_extract.csv")
        data = pd.read_csv(data_us_car_path, engine="python")
        return data

    else:
        raise ValueError("Dataset not found. Check the docstring for available values")



"""
Utils is a group of function for the library
"""

from pathlib import Path

import pandas as pd
from pandas.api.types import is_datetime64_any_dtype as is_datetime


def convert_string_to_int_keys(input_dict: dict) -> dict:
    """
    Returns the dict with integer keys instead of string keys
    Parameters
    ----------
    input_dict: dict
    Returns
    -------
    dict
    """
    return {int(k): v for k, v in input_dict.items()}


def base_100(series: pd.Series) -> pd.Series:
    """
    base_100 function put a pd.Series in base 100
    Parameters
    ----------
    serie: pd.Series
       input series to convert to base 100
    Returns
    -------
    pd.Series
        converted series
    """
    tot = series.sum()
    return series / tot


def get_project_root():
    """
    Returns project root absolute path.
    """
    current_path = Path(__file__)
    return current_path.parent.parent.resolve()


def truncate_str(text, maxlen=40):
    """
    truncate a string
    Parameters
    ----------
    text : string
        string to check in order to add line break
    maxlen : int
        number of characters before truncation
    Returns
    -------
    string
        truncated text
    """
    if isinstance(text, str) and len(text) > maxlen:
        tot_length = 0
        input_words = text.split()
        output_words = []
        for word in input_words[:-1]:
            tot_length = tot_length + len(word)
            if tot_length <= maxlen:
                output_words.append(word)

        text = " ".join(output_words)
        if len(input_words) > len(output_words):
            text = text + "..."
    return text


def round_to_k(x, k):
    """
    round float to k significant figure
    Parameters
    ----------
    x : float
        number to round
    k : int
        the number of significant figures
    Returns
    -------
    float or int
    """
    x = float(x)
    new_x = float("%s" % float(f"%.{k}g" % x))  # Rounding to k important figures
    if new_x % 1 == 0:
        return int(new_x)  # Avoid the '.0' that can mislead the user that it may be a round number
    else:
        return new_x


def convert_date_col_into_multiple_col(df: pd.DataFrame) -> pd.DataFrame:
    """
    Transform datetime column into multiple columns
        - year
        - month
        - day
    Drop datetime column
    Parameters
    ----------
    df: pd.Dataframe
       input DataFrame with datetime columns
    Returns
    -------
    pd.Dataframe
        DataFrame without datetime columns
    """

    date_col_list = [column for column in df.columns if is_datetime(df[column])]

    for col_date in date_col_list:
        df[col_date + "_year"] = df[col_date].dt.year
        df[col_date + "_month"] = df[col_date].dt.month
        df[col_date + "_day"] = df[col_date].dt.day

        # droping original date column
        df = df.drop(col_date, axis=1)

    return df



"""
This script can be used to generate the report example.
For more information, please refer to the tutorial 'tutorial01-modeldrift.ipynb' in tutorial/model_drift
that generates the same report.
"""
import os
import sys
from pathlib import Path

import catboost
import pandas as pd
from category_encoders import OrdinalEncoder
from sklearn import metrics
from sklearn.model_selection import train_test_split

sys.path.insert(0, "../..")

from eurybia import SmartDrift
from eurybia.data.data_loader import data_loading

if __name__ == "__main__":
    cur_dir = os.path.dirname(os.path.abspath(__file__))

    df_car_accident = data_loading("us_car_accident")
    df_accident_baseline = df_car_accident.loc[df_car_accident["year_acc"] == 2016]
    df_accident_2017 = df_car_accident.loc[df_car_accident["year_acc"] == 2017]
    df_accident_2018 = df_car_accident.loc[df_car_accident["year_acc"] == 2018]
    df_accident_2019 = df_car_accident.loc[df_car_accident["year_acc"] == 2019]
    df_accident_2020 = df_car_accident.loc[df_car_accident["year_acc"] == 2020]
    df_accident_2021 = df_car_accident.loc[df_car_accident["year_acc"] == 2021]

    y_df_learning = df_accident_baseline["target"].to_frame()
    X_df_learning = df_accident_baseline[
        df_accident_baseline.columns.difference(["target", "target_multi", "year_acc", "Description"])
    ]

    y_df_2017 = df_accident_2017["target"].to_frame()
    X_df_2017 = df_accident_2017[
        df_accident_2017.columns.difference(["target", "target_multi", "year_acc", "Description"])
    ]

    y_df_2018 = df_accident_2018["target"].to_frame()
    X_df_2018 = df_accident_2018[
        df_accident_2018.columns.difference(["target", "target_multi", "year_acc", "Description"])
    ]

    y_df_2019 = df_accident_2019["target"].to_frame()
    X_df_2019 = df_accident_2019[
        df_accident_2019.columns.difference(["target", "target_multi", "year_acc", "Description"])
    ]

    y_df_2020 = df_accident_2020["target"].to_frame()
    X_df_2020 = df_accident_2020[
        df_accident_2020.columns.difference(["target", "target_multi", "year_acc", "Description"])
    ]

    y_df_2021 = df_accident_2021["target"].to_frame()
    X_df_2021 = df_accident_2021[
        df_accident_2021.columns.difference(["target", "target_multi", "year_acc", "Description"])
    ]

    features = [
        "Start_Lat",
        "Start_Lng",
        "Distance(mi)",
        "Temperature(F)",
        "Humidity(%)",
        "Visibility(mi)",
        "day_of_week_acc",
        "Nautical_Twilight",
        "season_acc",
    ]

    features_to_encode = [
        col for col in X_df_learning[features].columns if X_df_learning[col].dtype not in ("float64", "int64")
    ]
    encoder = OrdinalEncoder(cols=features_to_encode)
    encoder = encoder.fit(X_df_learning[features])
    X_df_learning_encoded = encoder.transform(X_df_learning)

    Xtrain, Xtest, ytrain, ytest = train_test_split(
        X_df_learning_encoded, y_df_learning, train_size=0.75, random_state=1
    )
    train_pool_cat = catboost.Pool(data=Xtrain, label=ytrain, cat_features=features_to_encode)
    test_pool_cat = catboost.Pool(data=Xtest, label=ytest, cat_features=features_to_encode)

    model = catboost.CatBoostClassifier(
        loss_function="Logloss",
        eval_metric="Logloss",
        learning_rate=0.143852,
        iterations=300,
        l2_leaf_reg=15,
        max_depth=4,
        use_best_model=True,
        custom_loss=["Accuracy", "AUC", "Logloss"],
    )

    model = model.fit(train_pool_cat, plot=False, eval_set=test_pool_cat, verbose=False)
    SD = SmartDrift(df_current=X_df_2017, df_baseline=X_df_learning, deployed_model=model, encoding=encoder)
    SD.compile(
        full_validation=True,
        date_compile_auc="01/01/2017",
        datadrift_file=os.path.join(cur_dir, "car_accident_auc.csv"),
    )

    proba = model.predict_proba(X_df_2017)
    performance = metrics.roc_auc_score(y_df_2017, proba[:, 1]).round(5)
    df_performance = pd.DataFrame({"annee": [2017], "mois": [1], "performance": [performance]})

    SD = SmartDrift(df_current=X_df_2018, df_baseline=X_df_learning, deployed_model=model, encoding=encoder)
    SD.compile(
        full_validation=True,
        date_compile_auc="01/01/2018",  # optionnal, by default date of compile
        datadrift_file=os.path.join(cur_dir, "car_accident_auc.csv"),
    )

    proba = model.predict_proba(X_df_2018)
    performance = metrics.roc_auc_score(y_df_2018, proba[:, 1]).round(5)
    df_performance = df_performance.append({"annee": 2018, "mois": 1, "performance": performance}, ignore_index=True)

    SD = SmartDrift(df_current=X_df_2019, df_baseline=X_df_learning, deployed_model=model, encoding=encoder)
    SD.compile(
        full_validation=True,
        date_compile_auc="01/01/2019",  # optionnal, by default date of compile
        datadrift_file=os.path.join(cur_dir, "car_accident_auc.csv"),
    )

    proba = model.predict_proba(X_df_2019)
    performance = metrics.roc_auc_score(y_df_2019, proba[:, 1]).round(5)
    df_performance = df_performance.append({"annee": 2019, "mois": 1, "performance": performance}, ignore_index=True)

    SD = SmartDrift(df_current=X_df_2020, df_baseline=X_df_learning, deployed_model=model, encoding=encoder)
    SD.compile(
        full_validation=True,
        date_compile_auc="01/01/2020",  # optionnal, by default date of compile
        datadrift_file=os.path.join(cur_dir, "car_accident_auc.csv"),
    )

    proba = model.predict_proba(X_df_2020)
    performance = metrics.roc_auc_score(y_df_2020, proba[:, 1]).round(5)
    df_performance = df_performance.append({"annee": 2020, "mois": 1, "performance": performance}, ignore_index=True)

    SD = SmartDrift(df_current=X_df_2021, df_baseline=X_df_learning, deployed_model=model, encoding=encoder)
    SD.compile(
        full_validation=True,
        date_compile_auc="01/01/2021",  # optionnal, by default date of compile
        datadrift_file=os.path.join(cur_dir, "car_accident_auc.csv"),
    )

    proba = model.predict_proba(X_df_2021)
    performance = metrics.roc_auc_score(y_df_2021, proba[:, 1]).round(5)
    df_performance = df_performance.append({"annee": 2021, "mois": 1, "performance": performance}, ignore_index=True)
    SD.add_data_modeldrift(dataset=df_performance, metric="performance")

    SD.generate_report(
        output_file=os.path.join(cur_dir, "report.html"),
        title_story="Model drift Report",
        title_description="""US Car accident model drift 2021""",
        project_info_file=os.path.join(
            Path(os.path.abspath(__file__)).parent.parent, "eurybia/data", "project_info_car_accident.yml"
        ),
    )
    os.remove(os.path.join(cur_dir, "car_accident_auc.csv"))



"""
data analysis module used to compute added analysis on report
"""

from typing import Optional

import pandas as pd

from eurybia.report.common import VarType, display_value, replace_dict_values
from eurybia.utils.utils import round_to_k


def perform_global_dataframe_analysis(df: Optional[pd.DataFrame]) -> dict:
    """
    Returns a python dict containing global information about a pandas DataFrame :
    Number of features, Number of observations, missing values...
    Parameters
    ----------
    df : pd.DataFrame
        The dataframe that will be used to compute global information.
    Returns
    -------
    global_d : dict
        dictionary that contains an ensemble of global information about the input dataframe.
    """
    if df is None:
        return dict()
    missing_values = df.isna().sum().sum()
    global_d = {
        "number of features": len(df.columns),
        "number of observations": df.shape[0],
        "missing values": missing_values,
        "% missing values": missing_values / (df.shape[0] * df.shape[1]),
    }

    for stat in global_d.keys():
        if stat == "number of observations":
            global_d[stat] = int(global_d[stat])  # Keeping the exact number
        elif isinstance(global_d[stat], float):
            global_d[stat] = round_to_k(global_d[stat], 3)

    replace_dict_values(global_d, display_value, ",", ".")

    return global_d


def perform_univariate_dataframe_analysis(df: Optional[pd.DataFrame], col_types: dict) -> dict:
    """
    Returns a python dict containing information about each column of a pandas DataFrame.
    The computed information depends on the type of the column.
    Parameters
    ----------
    df : pd.DataFrame
        The dataframe on which the analysis will be performed
    col_types : dict
        Dict of types for each column
    Returns
    -------
    d : dict
        A dict containing each column as keys and the corresponding
        dict of information for each column as values.
    """
    if df is None:
        return dict()
    d = df.describe().to_dict()
    for col in df.columns:
        if col_types[col] == VarType.TYPE_CAT:
            d[col] = {"distinct values": df[col].nunique(), "missing values": df[col].isna().sum()}

    for col in d.keys():
        for stat in d[col].keys():
            if stat in ["count", "distinct values"]:
                d[col][stat] = int(d[col][stat])  # Keeping the exact number here
            elif isinstance(d[col][stat], float):
                d[col][stat] = round_to_k(d[col][stat], 3)  # Rounding to 3 important figures

    replace_dict_values(d, display_value, ",", ".")

    return d
